library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
library UNISIM;
use UNISIM.VComponents.all;

-------------------------------------------------------------------------------------
-- implemented following the description in this paper: 
-- M.Kumm, P. Zipf: "Efficient High Speed Compression Trees on Xilinx FPGAs"
--https://pdfs.semanticscholar.org/0798/96e0061bfde69c78573fb8c326441c00a159.pdf
--there are two different versions that only differ in the number of input and output word width
-------------------------------------------------------------------------------------

entity four_to_two_opt is
generic(N,M : integer); -- N and M and width and height of the multiplier, can be substituted by a single variable
port(
    w : in std_logic_vector(N+M-3 downto 0); -- last stage can process two inputs more, 
    x : in std_logic_vector(N+M-3 downto 0); -- w and x can have one more position
    y : in std_logic_vector(N+M-4 downto 0);
    z : in std_logic_vector(N+M-4 downto 0);
    s : out std_logic_vector(N+M+1 downto 0); --last stage produces two times s but no c
    c : out std_logic_vector(N+M-4 downto 0)
);
end four_to_two_opt;

architecture Behavioral of four_to_two_opt is

signal lut_o : std_logic_vector(N+M-1 downto 0);
signal carry : std_logic_vector(N+M downto 0);
signal din : std_logic_vector(N+M-1 downto 0);

begin

din(N+M-1 downto N+M-2) <= (others => '0');	--DI for carry chain, highest two positions zero
din(N+M-3 downto 0) <= w(N+M-3 downto 0);	--DI for carry chain, input of last 4:2 compressor forwarded to the carry chain
lut_o(N+M-1 downto N+M-2) <= (others => '0'); 	--LUT output
carry(0) <= '0'; 				--initial carry

--configure every stage, last stage has a reduced numbers of inputs

GEN_COMP : 
for nr in 0 to N+M-3 generate --don't forget to change the amount of 4:2 compressors to generate 
    COMP : 
    if nr < N+M-3 generate
    COMPX : LUT6_2 
    generic map(
        INIT => X"000069960000E8E8"
    )
    port map(
        I0 => x(nr),
        I1 => y(nr),
        I2 => z(nr),
        I3 => w(nr),
        I4 => '0',
        I5 => '1',
        O5 => c(nr),
        O6 => lut_o(nr)
    );
    end generate COMP;
    COMPNM : 
    if nr = N+M-3 generate
    COMPNMX : LUT6_2
    generic map(
        INIT => X"0000000000000006"
    )
    port map(
        I0 => x(nr),
        I1 => w(nr),
        I2 => '0',
        I3 => '0',
        I4 => '0',
        I5 => '0',
        O6 => lut_o(nr)
    );
    end generate COMPNM;  
end generate GEN_COMP;

--configuration of the carry chain

GEN_CARRY : 
for nr in 0 to (N+M)/4-1 generate
CARRYX : CARRY4
port map(
    CO => carry(nr*4+4 downto nr*4+1),
    O => s(nr*4+3 downto nr*4),
    CI => carry(nr*4),
    CYINIT => '0',
    DI => din(nr*4+3 downto nr*4),
    S => lut_o(nr*4+3 downto nr*4)                
);
end generate GEN_CARRY;    

end Behavioral;
