library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

use IEEE.NUMERIC_STD.ALL;

library UNISIM;
use UNISIM.VComponents.all;

entity tree_booth_compressor_modified is
generic (N,M : integer); --N : width, M : height
port(
    a : in std_logic_vector(N-1 downto 0);
    b : in std_logic_vector(M-1 downto 0);
    p : out std_logic_vector(N+M-1 downto 0)
);
end tree_booth_compressor_modified;

architecture Behavioral of tree_booth_compressor_modified is

function f_log2 (x: positive) return natural is --function to calculate log2(x)
    variable i : natural;
begin
    i := 0;
    while (2**i < x) and i < 31 loop
        i := i +1;
    end loop;
    return i;
end function;

--lut_o(stage)(lut position in stage), output of all luts, summarized per stage
type tLut_o is array(M/2-1 downto 0) of std_logic_vector(N+7 downto 0);--std_logic_vector(N+3 downto 0);
signal lut_o : tLut_o;
--carry(stage)(bit position in stage), carry word per stage including the carry in at position carry(stage)(0)
type tCarry is array(M/2-1 downto 0) of std_logic_vector(N+8 downto 0);--std_logic_vector(N+4 downto 0);
signal carry : tCarry;
--ccout(stage)(bit position in stage), carry chain output word per stage
type tOut is array(M/2-1 downto 0) of std_logic_vector(N+7 downto 0);--std_logic_vector(N+3 downto 0);
signal ccout : tOut;
--din(stage)(lut input), signal that is propagated to carry chain, therefore additional word for generic carry chain generation
type tDin is array(M/2-1 downto 0) of std_logic_vector(N+7 downto 0);--std_logic_vector(N+3 downto 0);
signal din : tDin;
--row(stage)(bit position in stage), carry chain output of a row prepared for addition
type tRow is array(M-1 downto 0) of std_logic_vector(N+M+4 downto 0);
--one more level for compressor tree rows
type ttRow is array(f_log2(M)-2 downto 0) of tRow;
signal row : ttRow;

begin

GEN_LUT:
for stage in 0 to M/2-1 generate
    LUTA0: 
    if stage = 0 generate
    GEN_LUTA0:
    for nr in 0 to N+1 generate
        LUTA00:
        if nr = 0 generate
            LUTA00X: LUT6_2
            generic map(
                INIT => X"0CC00CC0066CCAA0" --generate the LUT producing pp(0,0) and pp(0,1)
            )
            port map(
                I0 => a(1),
                I1 => a(0),
                I2 => '0',
                I3 => b(2*stage),
                I4 => b(2*stage+1),
                I5 => '1',
                O5 => ccout(stage)(1),
                O6 => ccout(stage)(0)
            );
        end generate LUTA00;
        LUTA01:                  
        if nr = 1 generate 
            LUTA01X: LUT6_2
            generic map(
                INIT => X"0056563CCCAAAA00" --generate the LUT producing pp(0,2)
            )
            port map(
                I0 => a(2),
                I1 => a(1),
                I2 => a(0),
                I3 => '0',
                I4 => b(2*stage),
                I5 => b(2*stage+1),
                O6 => ccout(stage)(2)
            );
        end generate LUTA01;
        LUTA02:                  
        if nr = 2 generate 
            LUTA02X: LUT6_2
            generic map(
                INIT => X"0001010300000000" --generate the LUT to produce initial carry chain input
            )
            port map(
                I0 => a(2),
                I1 => a(1),
                I2 => a(0),
                I3 => '0',
                I4 => b(2*stage),
                I5 => b(2*stage+1),
                O6 => lut_o(stage)(nr+1)
            );
        end generate LUTA02;        
        LUTA0Y:
        if nr > 2 and nr < N generate
            LUTA0YX: LUT6_2
            generic map(
                INIT => X"000000000335ACC0" --regular LUT for Booth encoding in the first row
            )
            port map(
                I0 => a(nr-1),
                I1 => a(nr),
                I2 => '0',
                I3 => b(2*stage),
                I4 => b(2*stage+1),
                I5 => '0',
                O6 => lut_o(stage)(nr+1)
            );
        end generate LUTA0Y;      
        LUTA0N1:
        if nr = N generate
            LUTA0N1X: LUT6_2
            generic map(
                INIT => X"000000000335ACC0" --last Booth encoding LUT in the first row
            )
            port map(
                I0 => a(N-1),
                I1 => a(N-1),
                I2 => '0',
                I3 => b(2*stage),
                I4 => b(2*stage+1),
                I5 => '0',
                O6 => lut_o(stage)(nr+1)
            );
        end generate LUTA0N1;
        LUTB0N2:
        if nr = N+1 generate
            LUTB0N2X:LUT6_2   
            generic map(
                INIT => X"0000000056669995" --LUT required for sign handling
            )
            port map(
                I0 => '1',
                I1 => a(N-1),
                I2 => '0',
                I3 => b(2*stage),
                I4 => b(2*stage+1),
                I5 => '0',
                O6 => lut_o(stage)(nr+1)
            );
        end generate LUTB0N2;
    end generate GEN_LUTA0;
    end generate LUTA0;
    
    LUTAY: 
    if stage > 0 and stage < M/2 generate
    GEN_LUTAY:
    for nr in 0 to N+1 generate
        LUTAY0:
        if nr = 0 generate
            LUTAY0X: LUT6_2
            generic map(
                INIT => X"0CC00CC0066CCAA0" --generate the LUT producing pp(N,0) and pp(N,1)
            )
            port map(
                I0 => a(1),
                I1 => a(0),
                I2 => b(2*stage-1),
                I3 => b(2*stage),
                I4 => b(2*stage+1),
                I5 => '1',
                O5 => ccout(stage)(1),
                O6 => ccout(stage)(0)
            );
        end generate LUTAY0; 
        LUTAY1:                 
        if nr = 1 generate 
            LUTAY1X: LUT6_2
            generic map(
                INIT => X"0056563CCCAAAA00" --generate the LUT producing pp(N,2)
            )
            port map(
                I0 => a(2),
                I1 => a(1),
                I2 => a(0),
                I3 => b(2*stage-1),
                I4 => b(2*stage),
                I5 => b(2*stage+1),
                O6 => ccout(stage)(2)
            );
        end generate LUTAY1;
        LUTAY2:                 
        if nr = 2 generate 
            LUTAY2X: LUT6_2
            generic map(
                INIT => X"0001010300000000" --generate the LUT to produce initial carry chain input
            )
            port map(
                I0 => a(2),
                I1 => a(1),
                I2 => a(0),
                I3 => b(2*stage-1),
                I4 => b(2*stage),
                I5 => b(2*stage+1),
                O6 => lut_o(stage)(nr+1)
            );     
        end generate LUTAY2;        
        LUTAYY:
        if nr > 2 and nr < N generate
            LUTAYYX: LUT6_2
            generic map(
                INIT => X"000000000335ACC0" --regular LUT for Booth encoding
            )
            port map(
                I0 => a(nr-1),
                I1 => a(nr),
                I2 => b(2*stage-1),
                I3 => b(2*stage),
                I4 => b(2*stage+1),
                I5 => '0',
                O6 => lut_o(stage)(nr+1)
            );
        end generate LUTAYY;
        LUTAYN1:
        if nr = N generate
            LUTAYN1X: LUT6_2
            generic map(
                INIT => X"000000000335ACC0" --last Booth encoding LUT
            )
            port map(
                I0 => a(N-1),
                I1 => a(N-1),
                I2 => b(2*stage-1),
                I3 => b(2*stage),
                I4 => b(2*stage+1),
                I5 => '0',
                O6 => lut_o(stage)(nr+1)
            );
        end generate LUTAYN1;
        LUTBYN2:
        if nr = N+1 generate
            LUTBYN2X:LUT6_2       
            generic map(
                INIT => X"0000000056669995" --LUT required for sign handling
            )
            port map(
                I0 => '0',
                I1 => a(N-1),
                I2 => b(2*stage-1),
                I3 => b(2*stage),
                I4 => b(2*stage+1),
                I5 => '0',
                O6 => lut_o(stage)(nr+1)
            );
        end generate LUTBYN2;
    end generate GEN_LUTAY;
    end generate LUTAY;
    
end generate GEN_LUT;

--generate the carry chain
GEN_CARRY :
for stage in 0 to M/2-1 generate
    GEN_CARRYY : 
    for nr in 0 to (N+4)/4-1 generate
    CARRYYX : CARRY4
    port map(
        CO => carry(stage) (nr*4+4+3 downto nr*4+1+3),
        O => ccout(stage) (nr*4+3+3 downto nr*4+3),
        CI => carry(stage) (nr*4+3),
        CYINIT => '0',
        DI => din(stage) (nr*4+3+3 downto nr*4+3),
        S => lut_o(stage) (nr*4+3+3 downto nr*4+3)                
    );
    end generate GEN_CARRYY;    
end generate GEN_CARRY;  

--summarize all the outputs in partial product rows to have a simplified handling
OUTPUT:
for i in 0 to (M/2)-1 generate
OUTPUT_1:
if i = 0 generate
    carry(i)(3) <= '1';
    lut_o(i)(N+3) <= '1';
    din(i)(N+3 downto 0) <= (b"01" & (N+1 downto 0 => '0'));
    row(0)(i)(N+M-1 downto 0) <= (((N+M)-(N+4)-1 downto 0 => '0') & carry(i)(N+4) & ccout(i)(N+3 downto 4) & ccout(i)(2 downto 0));
end generate OUTPUT_1;
OUTPUT_Y:
if i > 0 and i < (M/2)-1 generate
    carry(i)(3) <= '1';
    lut_o(i)(N+3) <= '1';
    din(i)(N+3 downto 0) <= (others => '0');
    row(0)(i)(N+M-1 downto 0) <= (((N+M)-(N+4)-(i*2)-1 downto 0 => '0') & carry(i)(N+4) & ccout(i)(N+3 downto 4) & ccout(i)(2 downto 0) & ((2*i)-1 downto 0 => '0'));
end generate OUTPUT_Y;   
OUTPUT_M:
if i = (M/2)-1 generate
    carry(i)(3) <= '1';
    din(i)(N+3 downto 0) <= (others => '0');
    row(0)(i)(N+M-1 downto 0) <= (ccout(i)(N+2 downto 4) & ccout(i)(2 downto 0) & ((2*i)-1 downto 0 => '0'));
end generate OUTPUT_M; 
end generate OUTPUT;

--generate the required number of 4:2 compressors
GEN_COM:
for level in 0 to (f_log2(M)-3) generate --level is required since for bigger multiplier several compressor levels are required until only 2 rows are remaining for RCA
GEN_COM_NR:
for nr in 0 to (M/((level+1)*8)-1) generate --generates a long "row" of 4:2 compressors
compressor : entity work.four_to_two_opt
generic map(
    N => N,
    M => M
)
port map(
    w => row(level)((nr*4)+3)(N+M-1 downto 2),
    x => row(level)((nr*4)+2)(N+M-1 downto 2),
    y => row(level)((nr*4)+1)(N+M-2 downto 2),
    z => row(level)((nr*4))(N+M-2 downto 2),
    s => row(level+1)(nr*2)(N+M+3 downto 2),
    c => row(level+1)((nr*2)+1)(N+M-1 downto 3)
);
row(level+1)(nr*2)(1 downto 0) <= row(level)(nr*2)(1 downto 0);
row(level+1)((nr*2)+1)(1 downto 0) <=(others => '0');
row(level+1)((nr*2)+1)(2) <= '0';
end generate GEN_COM_NR;
end generate GEN_COM;

--generate the final ripple carry adder to produce the final product
rca : entity work.rca
generic map(
    N => N+M
)
port map(
    a => row(f_log2(M)-2)(0)(N+M-1 downto 0),
    b => row(f_log2(M)-2)(1)(N+M-1 downto 0),
    c_in => '0',
    s => p
);


------------------------------------------------------------------------------------------
--Input & output configuration for 16x16, example for manual compressor instantiation
------------------------------------------------------------------------------------------


--compressor00 : entity work.four_to_two
--generic map(
--    N => N,
--    M => M
--)
--port map(
--    w => row(0)(0)(N+M downto 0),
--    x => row(0)(1)(N+M downto 0),
--    y => row(0)(2)(N+M-1 downto 0),
--    z => row(0)(3)(N+M-1 downto 0),
--    s => row(1)(0)(N+M+4 downto 0),
--    c => row(1)(1)(N+M downto 1)
--);
--row(1)(1)(N+M+4 downto N+M+1) <= b"0000";
--row(1)(1)(0) <= '0';

--compressor01 : entity work.four_to_two
--generic map(
--    N => N,
--    M => M
--)
--port map(
--    w => row(0)(4)(N+M downto 0),
--    x => row(0)(5)(N+M downto 0),
--    y => row(0)(6)(N+M-1 downto 0),
--    z => row(0)(7)(N+M-1 downto 0),
--    s => row(1)(2)(N+M+4 downto 0),
--    c => row(1)(3)(N+M downto 1)
--);
--row(1)(3)(N+M+4 downto N+M+1) <= b"0000";
--row(1)(3)(0) <= '0';

--compressor10 : entity work.four_to_two
--generic map(
--    N => N,
--    M => M
--)
--port map(
--    w => row(1)(0)(N+M downto 0),
--    x => row(1)(1)(N+M downto 0),
--    y => row(1)(2)(N+M-1 downto 0),
--    z => row(1)(3)(N+M-1 downto 0),
--    s => row(2)(0)(N+M+4 downto 0),
--    c => row(2)(1)(N+M downto 1)
--);
--row(2)(1)(N+M+4 downto N+M+1) <= b"0000";
--row(2)(1)(0) <= '0';

--rca20 : entity work.rca
--generic map(
--    N => N+M
--)
--port map(
--    a => row(2)(0)(N+M-1 downto 0),
--    b => row(2)(1)(N+M-1 downto 0),
--    c_in => '0',
--    s => p
--);

------------------------------------------------------------------------------------------
    
end Behavioral;
